@@ -394,6 +394,9 @@ xf86WaitForInput(int fd, int timeout)
     poll_fd.fd = fd;
     poll_fd.events = POLLIN;
 
+    /* convert microseconds to milliseconds */
+    timeout = (timeout + 999) / 1000;
+
     if (fd >= 0) {
         SYSCALL(r = xserver_poll(&poll_fd, 1, timeout));
     }
@@ -482,7 +485,7 @@ static struct states {
 #endif
 };
 
-static int numStates = sizeof(modemStates) / sizeof(modemStates[0]);
+static int numStates = ARRAY_SIZE(modemStates);
 
 static int
 xf2osState(int state)
